#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

#program to analyze blat results and extract the corresponding genomics regoin from 2bit file

use strict;
use warnings;
use Readonly;
use Carp;
use POSIX qw (ceil);
use Storable;
use Pod::Usage;

our $VERSION = 1.10;

if (@ARGV < 5) {usage()};

#get user parameters:
my $blat_file           = shift;
    # blat file should be in pslx format. it may contain the header lines.
my $threshold           = shift;
    # number.
my $number_of_cycles    = shift;

my $genome_file         = shift;
    # only 2bit version 0 format. little and big endian  supported
my $outfile             = shift;
    # store hash table of position to sequence in Storable format.


#the actual actions:
my $r_best_blat_result_of =    read_pslx($blat_file,$threshold);
my $r_tree                =    get_2bit_tree($genome_file);
my $r_pos_to_seq          =    getDNA($r_best_blat_result_of,$r_tree,$number_of_cycles);

#r_tree->{FH} holds the filehandle of the 2bit genome sequnece
close $r_tree->{FH} 
    or croak "can't close";

# save the outputs
store($r_pos_to_seq, $outfile)
    or croak "can't store file\n";

print "+Number of lines:",scalar keys %$r_pos_to_seq,"\n";
exit;


sub usage {
    print "Usage:analyze_blat_neat.pl <blat_file> <threshold> <genome_file> <outfile>\n";
    croak "+analyze_blat_neat is dead\n";
}

sub getDNA {

    #############################################
    #this function takes blat hash from <read_pslx> function and return a hash table 
    # with the names of the queries and the matching genome sequences.
    # the function assumes ';' to seperate collapsed query names!!
    # It filters out Ns and return the minus strand if needed (rev comp). 
    # the sequences are in capital letters.
    # I used a random access approach (not grouping by names or sorting by offsets)
    # as from benchmarking tests it seems to work the best.
    #############################################


    my ($r_best_blat_result_of,$r_two_bit_tree,$number_of_cycles) = @_;
        #$number of cycles - how much do we want to read if we have a valid alignment.
    my %pos_to_seq;
        # stores the results


    GET_SEQ: foreach my $qName (keys %$r_best_blat_result_of) {

        # find the sequence name and strand
        my $tName  = $r_best_blat_result_of->{$qName}->{tName};
        my $strand = $r_best_blat_result_of->{$qName}->{strand};
        my $dna = undef;
            # $dna holds the sequence

        if ($strand eq '+') {
            # loc is in bases:
            my $loc =   $r_best_blat_result_of->{$qName}->{tStart} 
                      - $r_best_blat_result_of->{$qName}->{qStart}
                      + 1; # +1 to adapt to blat format. 
            next if (not ($dna = _fetch_DNA($tName,$loc,$number_of_cycles,$r_two_bit_tree)));
                #next in case of seqences that are at the end. (see "safety" in _fetch_DNA function).
        }
        else {
            # loc is in bases:
            my $loc =    $r_best_blat_result_of->{$qName}->{tEnd}
                       + $r_best_blat_result_of->{$qName}->{qStart}
                       - $number_of_cycles
                       + 1; # +1 to adapt to blat format. 
            next if (not ($dna = _fetch_DNA($tName,$loc,$number_of_cycles,$r_two_bit_tree)));
                #next in case of seqences that are at the end. (see "safety" in _fetch_DNA function).
            $dna =~ tr/ACGT/TGCA/;
            $dna = reverse $dna;
        }

        $dna = uc $dna;
        next if ($dna =~ m/N/);
            # filter out sequence with N.

        my @names = split /;/, $qName;
        foreach my $name (@names) {
            $pos_to_seq{$name} = $dna;
        }
    } # end of GET_SEQ

    return \%pos_to_seq;

}



sub read_pslx {

    #############################################
    # This function takes pslx file and a threshold and gives it returns the file in a hash table format.
    # The pslx file may contain headers. threshold is a number. The function filters out alignment with
    # inserts (rarely happend in Solexa), with Ns, below threshold, or non unique alignments. Non unique
    # alignments refers to a situtation when the qSequence (the query sequence) matches with the same best 
    # score to more than one region. We want unique alignment for the training to avoid any ambig.
    # The return hash table does not hold all the file, but important fields. Not all of them are 
    # processed later, but just in case...
    # some variable names anf their description were taken from BLAT by WJ Kent.
    ###############################################

    my ($pslx_file, $threshold) = @_;
    open my $PSLX, '<', $pslx_file
        or croak "Can't open pslx file for reading\n";
    my %best_blat_result_of;


    READ_BLAT: while (<$PSLX>) {
	next if (not m/^\d/); #filter header
	chomp;
	my @line = split /\s+/ , $_;
        # load the fields:
        my $matches        = $line[0];        # Number of bases that match that aren't repeats
        my $misMatches     = $line[1];        # Number of bases that don't match
        my $repMatches     = $line[2];        # Number of bases that match but are part of repeats
        my $nCount         = $line[3];        # Number of 'N' bases
        my $qNumInsert     = $line[4];        # Number of inserts in query
        my $qBaseInsert    = $line[5];        # Number of bases inserted in query
        my $tNumInsert     = $line[6];        # Number of inserts in target
        my $tBaseInsert    = $line[7];        # Number of bases inserted in target
        my $strand         = $line[8];        # + or - for query strand
        my $qName          = $line[9];        # Query sequence name
        my $qSize          = $line[10];       # Query sequence size
        my $qStart         = $line[11];       # Alignment start position in query
        my $qEnd           = $line[12];       # Alignment end position in query
        my $tName          = $line[13];       # Target sequence name
        my $tSize          = $line[14];       # Target sequence size
        my $tStart         = $line[15];       # Alignment start position in target
        my $tEnd           = $line[16];       # Alignment end position in target
        my $blockCount     = $line[17];       # Number of blocks in alignment. A block contains no gaps.
        my $blockSizes     = $line[18];       # Size of each block in a comma separated list
        my $qStarts        = $line[19];       # Start of each block in query in a comma separated list
        my $tStarts        = $line[20];       # Start of each block in target in a comma separated list

        my $qSequence      = $line[21];       # Block of query sequence
        my $tSequence      = $line[22];       # Block of target sequence

        #filter alignments with Ns, non continuous, or below thershold:
        next if ( ($matches < $threshold) or 
                  ($tBaseInsert > 0)      or 
                  ($nCount      > 0)      or
                  ($qBaseInsert > 0)    );

        #mark non unique alignments. We will delete them later
        if ( (exists $best_blat_result_of{$qName}) and 
             ($best_blat_result_of{$qName}{'matches'} == $matches)
           ) {
             $best_blat_result_of{$qName}{'NotUnique'} = 1;
        }

        # if a new alignment or better alignment arrives, store it:
        if ( (not exists $best_blat_result_of{$qName}) ||
			 ($best_blat_result_of{$qName}{'matches'} < $matches) ) {

            $best_blat_result_of{$qName}{'matches'}  = $matches  ;
            $best_blat_result_of{$qName}{'strand'}   = $strand   ;
            $best_blat_result_of{$qName}{'tName'}    = $tName    ;
            $best_blat_result_of{$qName}{'qStart'}   = $qStart   ;
            $best_blat_result_of{$qName}{'tStart'}   = $tStart   ;
            $best_blat_result_of{$qName}{'tEnd'}     = $tEnd     ;
            $best_blat_result_of{$qName}{'tSequence'}= $tSequence;	 
            $best_blat_result_of{$qName}{'NotUnique'} = 0        ;

        }


    } #end of READ_BLAT

    # get rid of non unique alignments
    foreach my $qName (keys %best_blat_result_of) {
        if ($best_blat_result_of{$qName}{'NotUnique'} == 1) {
            delete $best_blat_result_of{$qName};
        }
    }

    close $PSLX;

    return \%best_blat_result_of;
}



sub _fetch_DNA {

    # get a name of a sequence. a location in basepairs and a length to etract a sequence.
    # the function reports back 'N's BUT DOES NOT MASK (mblocks!). THE FUNCTION DOES NOT 
    # DEAL WITH QUERIES THAT ENDS UP LESS THAN 5 BASES BELOW THE SEQUENCE ENDS. 
    # it returns the plus strand sequence and not define in a case of failure.

    Readonly my $SAFETY => '5'; # in bases 
        # since we are rounding up problems can occur.
        # it might be over safety (because of padding), but it prevernts fatal errors.
        # next versions will deal with that.

    my ($name,$location,$length,$r_tree) = @_;
    my $TWO_BIT = $r_tree->{FH};
    my $template = $r_tree->{template};
    my $r_l;
    my $raw;

    # set zero as UCSC browser. This is not like BLAT results!
    $location--;

    # prevent queries that reach the end of the sequence. 
    if ($r_tree->{name}->{$name}->{lengthDNA} < ($location + $length) + $SAFETY) {
        return;
    }

    my $dna_offset_start   = $r_tree->{name}->{$name}->{offsetDNA};
        # from the begining of the file
    my $offset_bytes       = int ($location / 4) + $dna_offset_start;
        # convert bases to bytes and round down. the begining of the sequence we want to read.
    my $reminder_start     = $location % 4 ;
        # number of extra bases that were added to the 5' end
    my $len_bytes          = ceil ($length / 4) + 1;
        # nubmer of bytes to read. checked with UCSC browser.
    my $end                = $location + $length;
        # end is in bases

    # go to the sequence and read:
    seek $TWO_BIT, $offset_bytes, 0
        or croak "problem in _fetch_DNA\n";
    $r_l = read $TWO_BIT, $raw, $len_bytes;
    _check_read($r_l, $len_bytes);
        #always check that you get all the bytes that you wanted to read!

    # use the lookup array to convert the bytes to sequence:
    my $dna = join '', map{ ${$r_tree->{'map'}}[ $_ ] } unpack 'C*', $raw;
    $dna = substr $dna, $reminder_start , $length ;
        # get rid of the extra nucleotides on the 5'.

    ###
    # 2bit format has a funny way to represent 'N's. The Nblock fields in the 
    # sequence detail header holds the starting points. IMPORTANT: the file includes 
    # the sequence of the 'N's (usually as 'T's), so one needs to REPLACE and NOT to INSERT.
    # I checked with UCSC browser and the function knows how to deals with Ns in the middle 
    # of the sequence to fetch or in the begining (in the middle of the block).

    # for each starting poing of nblock (in bases):
    foreach my $nblock_start (@{$r_tree->{name}->{$name}->{nblock}->{start}}) {

        # get the end of the nblock (in bases):
         my $nblock_end = $r_tree->{name}->{$name}->{nblock}->{len}->{$nblock_start} + $nblock_start;

         next if ($location > $nblock_end);
         last if ($end < $nblock_start);
            # since the starting points are sorted we can stop searching
            # when the first block starting point occurs after the end 
            # of the sequence to fetch

        #
        #from here only blocks that occur in the overlap with the sequence!!
        #

         # calculate the relative points to the fetched sequence (in bases):
         my $nblock_start_loc = $nblock_start - $location;
         my $nblock_end_loc   = $nblock_end   - $location;

         # get only the overlaped positions:
         if ($nblock_start_loc < 0)     {$nblock_start_loc = 0};
         if ($nblock_end_loc > $length) {$nblock_end_loc = $length};
         my $length_to_fix = $nblock_end_loc - $nblock_start_loc;

         # replace the sequence with Ns:
         substr $dna, $nblock_start_loc, $length_to_fix, 'N' x $length_to_fix;
    }

    return $dna;

}

sub get_2bit_tree {

    #############################################
    #This function takes 2bit genome format and extracts a tree
    # that describes the file. The tree contains: (1) endianess of the file 
    # (2) filehandle to the 2bit file (3) names and locations of the DNA sequences
    # more infomration in: http://genome.ucsc.edu/FAQ/FAQformat#format7. This 
    # function was built with the insperation from: http://www.perlmonks.org/?node_id=672251
    # The tree:
    # $tree{FH}                                 => filehandle to 2bit file
    # $tree{template}                           => either N or V for endianess
    # $tree{map}                                => array to convert between a byte and a sequence
    # $tree{name}                               => data regarding each DNA sequence.
    # $tree{'name'}{$name}{'offsetHeader'}      => offset to details about the sequence (and the later the sequence).
    #                                               from the begining of the file
    # $tree{name}{$name}{offsetDNA}             => the offset of sequence <$name> from the begining of the file
    # $tree{name}{$name}{lengthDNA}             => number of bases of the sequence <$name>
    # @{$tree{name}{$name}{$block}{start}}      => sorted array of blocks starting points of sequence <$name>
    #                                               <$block> can be either 'nblock' or 'mblock'.
    # $tree{name}{$name}{$block}{len}{$start}   => the length of a block with a starting point <$start>.
    #############################################

    # load user parameters:
    my ($two_bit_file) = @_;
    my %tree;
    open my $TWO_BIT, '<', $two_bit_file
        or croak "can't open 2bit genome file: $!\n";

    # read the header of the file. Check endianess, version =0, and that it is 2bit file.
    # if it is report the number of sequences in te file and the endianess
    my ($num_of_seq, $template) = _read_2bit_header($TWO_BIT);
            #template is the endianess

    # store the filehandle and the endianess for subsequent functions:
    $tree{FH} = $TWO_BIT;
    $tree{template} = $template;

    # for each sequence read sequence header. This gives back the name of the sequence and its offset.
    # actually, the offset is for sequence details
    foreach my $seq_header (1..$num_of_seq) {
        _read_2bit_seqheader (\%tree);
    }

    # for each sequence read sequence details. The detials include the length of the bases, N blocks, 
    # repeat masking blocks and the offset from the begining of the file to the sequence.
    foreach my $seq_name (keys %{$tree{name}}) {
        _read_2bit_sequence_record($seq_name,\%tree);
    }

    # create a fast lookup array that converts between one byte to 4 nucleotides
    my %bitMap  =  ('00' => 'T', '01' => 'C', '10' => 'A', '11' => 'G');
    @{$tree{'map'}} = map{ 
                            join '', map $bitMap{ $_ }, unpack '(A2)4', unpack 'B8', chr 
                            } 0 .. 255;

    return \%tree;
}

sub _read_2bit_header {

    Readonly my $HEADER_LENGTH             => '16'       ;
    Readonly my $HEADER_SIGNATURE          => 0x1A412743 ;


    # read and prepare variables
    my ($TWO_BIT) = @_;
    my $r_l = 0;
        #r_l (read length) checks the actual number of bytes that were read
    my $raw = undef;
        #raw stroes the hex information from the file

    # reading the header:
    $r_l = read $TWO_BIT, $raw, $HEADER_LENGTH;
    _check_read($r_l, $HEADER_LENGTH);
        #always check that you get all the bytes that you wanted to read!

    # try to endianess and see which one is the correct one
    ENDIANESS: foreach my $template ("N", "V") {
        my ($signature, $version  ,
             $seq_count,$reserved ) = unpack $template.'4', $raw;
                    #there are the four fields. each one is 32bit long

        #version should always be zero!:
        if($signature==$HEADER_SIGNATURE and $version==0) 
            {return ($seq_count,$template)}
    } # end of ENDIANESS

    #If the signature does not match to anything or version is not zero:
    croak "File is not in a 2bit format or version is not zero\n";
}
 

sub _read_2bit_seqheader {

    Readonly my $NAME_SIZE_LENGTH => '1'; # in bytes
    Readonly my $OFFSET_LENGTH    => '4'; # in bytes

    # we have sequence headers as the same number of sequences.
    # the header holds: <length of name in bytes - 8bit> <name> <
    # offset to sequence details - 32bit>

    # read and prepare variables
    my ($r_tree) = @_;
    my $TWO_BIT = $r_tree->{FH};
    my $template = $r_tree->{template};
    my $r_l = 0;
    my $raw = undef;
    my $name = undef;

    # read name length:
    $r_l = read $TWO_BIT, $raw, $NAME_SIZE_LENGTH;
    _check_read($r_l, $NAME_SIZE_LENGTH);
        #always check that you get all the bytes that you wanted to read!
    my $nameSize = unpack 'C', $raw;

    # read name:
    $r_l = read $TWO_BIT, $name , $nameSize;
    _check_read($r_l,$nameSize);

    # read offset:
    $r_l = read $TWO_BIT, $raw, $OFFSET_LENGTH;
    _check_read($r_l,$OFFSET_LENGTH);
    my $offset = unpack $template, $raw;

    # store data:
    $r_tree->{'name'}->{$name}->{'offsetHeader'}     = $offset;

}


sub _read_2bit_sequence_record {

    #takes name of sequence and the tree and extract the sequence details:
    #length, number of N and M blocks, thier strating potins, size of each block
    #, and offset to the DNA sequence from the beging of the file.

    #I used the code of limbic region from perlmonks 
    Readonly my $DNA_SIZE_LENGTH     => '4'; # in bytes
    Readonly my $BLOCK_COUNT_LENGTH  => '4'; # in bytes
    Readonly my $BLOCK_STARTS_LENGTH => '4'; # in bytes
    Readonly my $BLOCK_SIZES_LENGTH  => '4'; # in bytes
    Readonly my $RESERVED_LENGTH     => '4'; # in bytes

    # read and prepare variables
    my ($name,$r_tree) = @_;
    my $TWO_BIT = $r_tree->{FH};
    my $template = $r_tree->{template};
    my $dna_length = 0;
    my $cnt;
    my (@starts, @lens);
    my $r_l;
    my $raw;

    # go to sequence details:
    my $offset = $r_tree->{'name'}->{$name}->{offsetHeader};
    seek $TWO_BIT, $offset,0 or croak "getDNA function error: $!\n";

    # get length of the sequence (total number of bases):
    $r_l = read $TWO_BIT, $raw, $DNA_SIZE_LENGTH;
    _check_read($r_l,$DNA_SIZE_LENGTH);
        #always check that you get all the bytes that you wanted to read!
    $dna_length = unpack $template,$raw;


    # get data regarding nblocks and mblocks ('N' nucs, and nucleotides to mask)
    foreach my $block ('nblock', 'mblock') {

        # read the number of blocks:
        $r_l = read $TWO_BIT, $raw, $BLOCK_COUNT_LENGTH ;
        _check_read($r_l,$BLOCK_COUNT_LENGTH);
        $cnt = unpack $template, $raw;
            # $cnt - how many blocks we have in the seqeunce

        # if there are blocks get their starting points and lengths:
        READ_BLOCKS: if ($cnt) {
            # read starting points:
            $r_l = read $TWO_BIT, $raw, $BLOCK_STARTS_LENGTH * $cnt;
            _check_read($r_l,$BLOCK_STARTS_LENGTH * $cnt);
            my $n_t = $template.$cnt;
            @starts = unpack $n_t, $raw;

            # read lengths:
            $r_l = read $TWO_BIT, $raw, $BLOCK_SIZES_LENGTH * $cnt;
            @lens = unpack $n_t, $raw;

            # store the starting points by order:
            @{$r_tree->{name}->{$name}->{$block}->{start}} = sort {$a <=> $b} @starts;
            # store the length in a hash format:
            foreach my $start (@starts) {
                my $len = shift @lens;
                $r_tree->{name}->{$name}->{$block}->{len}->{$start} = $len;
            }
        } # end of READ_BLOCKS
    }

    $r_tree->{name}->{$name}->{offsetDNA} = (tell $TWO_BIT) +  $RESERVED_LENGTH;
        # store the starting point of the sequence from the end of the file
    $r_tree->{name}->{$name}->{lengthDNA} = $dna_length;
        # store the length of the sequence (in bases)
}



sub _check_read {
    # aux function - see that we read as match as we wanted.
    # generates a fatal error if it is not correct
    my ($actual_l,$expected_l) = @_;
    if ($actual_l != $expected_l) {
        croak "problem to read file\n";
    }
    return 1;
}

__END__


=head1 NAME
    analyze_blat_neat.pl

=head1

=head1 VERSION
    1.0

=head1

=head1 SYNOPSIS
    Fetch genomic sequence from blat query. The primary aim is to extract sequences for training. Thus, sequences with Ns, not unique, or very noisy are not reported.

=head1

=head1 USEAGE
    analyze_blat_neat.pl <blat_file> <threshold> <number_of_cycles> <genome_file> <outfile>
    All fields are mandatory.

=head1

    <blat file>                     pslx format. BLAT header is allowed.
    <threshold>                     Sequences that their number of matches is lower than threshold will be filtered out.
    <number_of_cycles>              Number of cycles in the run. Determines the length of the DNA.
    <genome_file>                   2bit genome file. Little and big endian are supported.
    <outfile>                       Holds a hash file. In perl Storable format (binary). The hash is key:query nane -> genomic sequence.
                                    Query names can be in a collpased format with ';' for seperation.

=head1

=head1 DEPENDENCIES
    Readonly
    Carp
    Storable

=head1

=head1 BUGS AND LIMITATIONS
    The return sequences are not masked for repeats.
    If the best alignment occurs next to the end of the genome sequence (up to five bases) it will not be repored.
    If you feel you have found a bug in analyze_blat_neat.pl, please submit a bug report at erlich@cshl.edu

=head1

=head1 AUTHOR
    Written By Yaniv Erlich <erlich@cshl.edu>

=head1

=head1 LISCENSE AND COPYRIGHT
    Copyright © 2008 Cold Spring Harbor Laboratory.
    This is free software.  You may redistribute copies of it under the terms of the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.  There is NO WARRANTY, to the extent permitted by law.

=head1

=cut



