#! /usr/bin/env perl


#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

package Cluster::Remote_Scripts::Caller;


use strict;
use warnings;
use PDL;
use Readonly;
use Model3;
use List::Util qw (shuffle);
use Storable;
use File::Temp qw (tempfile);
use Carp;
use Getopt::Long;
use Archive::Tar;
use Cwd 'abs_path';
use File::Copy;
use File::Basename;



#Caller.pm --random_walk 1 2 --number_of_cycles 50 --intensity_file_to_call Y --verbose --svm-files --svm-scales --prob_estimate



Readonly my $SPACE              => q{ };
Readonly my $D_p3               => '0';
Readonly my $D_shrink           => '0';

Readonly my $D_SVM_SCALE_COMMAND    => 'svm-scale';
Readonly my $D_SVM_SCALE_PARAMS     => '-r';
Readonly my $D_SVM_PREDICT_COMMAND  => 'svm-predict';
Readonly my $D_SVM_PREDICT_PARAMS   => '';
Readonly my $EPS                    => '10e-10';
Readonly my $D_UPPER_LIMIT          => '40';
Readonly my $D_TMPELATE             => 'T_SVM_file_for_cycle_N_XXXX';
Readonly my $D_GUNZIP               => 'gunzip';
Readonly our $D_SUFFIX_CALLER        => '_AltaCyclic_seq.txt';

main();


sub main {

    my $time = time();
    print "+load parameters\n";
    my $r_config = read_parameters();

    print "+reporting parameters\n";
    report_params($r_config);

    print "+load_correction_matrix\n";
    my $deconv = load_correction_matrix ($r_config);

    print "+prepare_tmp_zone_for_calling\n";
    my @tmp_calling_files = prepare_tmp_zone_for_calling($r_config,$time);

    print "+arrange_SVM_files\n";
    my $r_SVM_of_cycle = arrange_SVM_files($r_config);

    print "+deconvolute intensity file\n";
    my ($counter,$r_count_to_pos) = deconv_int($deconv,$r_config);


    print "+close_tmp_zone_fh\n";
    close_tmp_zone_fh($r_config);


    print "+apply_scaling\n";
    my ($r_scaledfiles,$r_range_files) 
                = apply_scaling(\@tmp_calling_files,$r_SVM_of_cycle);

    print "+SVM_them\n";
    my $r_SVMed_files = SVM_them($r_scaledfiles,$r_SVM_of_cycle,$r_config);
    print "finished SVM train\n";

    print "+Seqencing\n";
    my $sequence_file = convert_to_sequence_and_probability ($r_SVMed_files,
                                                             $r_count_to_pos,
                                                             $r_config);

    print "+Sending Files\n";
    send_files_home($r_config->{bring_files_to},$sequence_file);

}



sub read_parameters {

    my @random_walk;
    my $intensity_file_to_call;
    my $number_of_cycles;
    my $prob_estimate;
    my $verbose;
    my $svm_files;
    my $bring_files_to;
    my $flowcell_name;

    my %config;



    GetOptions( "random_walk=f{2}"         => \@random_walk,
                "intensity_file_to_call=s" => \$intensity_file_to_call,
                "number_of_cycles=i"       => \$number_of_cycles,
                "prob_estimate"            => \$prob_estimate,
                "verbose"                  => \$verbose,
                "svm_files=s"              => \$svm_files,
                "bring_files_to=s"         => \$bring_files_to,
                "flowcell_name=s"          => \$flowcell_name,
                );


    my ($p1,$p2) = @random_walk;

    croak "<random_walk parameters> is not set\n"
        if (not defined $p1 or not defined $p2);

    croak "<number_of_cycles is> not set\n"
        if (not defined $number_of_cycles);

    croak "<intensity_file_to_call> is not set\n"
        if (not defined $intensity_file_to_call);

    croak "<svm_model_files> is not set\n"
        if (not defined $svm_files);

    croak "<bring files to> option should be set\n"
        if (not defined $bring_files_to);


    if ($verbose) {
        $config{'verbose'} = 1;
    } 
    else {
        $config{'verbose'} = undef;
    }


    if ($prob_estimate) {
        $prob_estimate = 'Yes';
    }
    else {
        $prob_estimate = 'No';
    }

    croak "<bring_files_to> argument must be absolute path\n"
        if (defined $bring_files_to and $bring_files_to=~m/^\./);

    $config{'p1'}                    = $p1;
    $config{'p2'}                    = $p2;
    $config{'number_of_cycles'}      = $number_of_cycles;
    $config{'intensity_file_to_call'}= $intensity_file_to_call;
    $config{'prob_estimate'}         = $prob_estimate;
    $config{'svm_files'}             = $svm_files;
    $config{'bring_files_to'}        = $bring_files_to;
    $config{'flowcell_name'}         = $flowcell_name;

    return \%config;
}

sub report_params {

    my ($r_config) = @_;
    my @params = qw (p1
                     p2
                     number_of_cycles
                     intensity_file_to_call
                     prob_estimate
                     svm_files
                     bring_files
                     );

    foreach my $key (@params) {

        my $v = $r_config->{$key} || 'undef';
        print ">","$key:","$v","\n";
    }
    return 1;
}

sub load_correction_matrix {

    my ($r_config)  =    @_;

    croak "load_correction_matrix error\n"
        if (not defined $r_config);

    my $end_cycle   =    $r_config->{'number_of_cycles'};
    my $p1          =    $r_config->{'p1'};
    my $p2          =    $r_config->{'p2'};

    my $m = Model3->new();
    $m->end_cycle($end_cycle);
    $m->set_parameters($p1,$p2,$D_p3,$D_shrink);
    $m->calculate_matrix();
    my (undef,undef,$deconv_matrix) = $m->inverse_polymerase_matrix();

    return $deconv_matrix;

}

sub prepare_tmp_zone_for_calling {

    my ($r_config,$time) = @_;


    my @tmpfilenames,
    my @fileHandles;
    my $number_of_cycles = $r_config->{number_of_cycles};

    foreach my $cycle (1..$number_of_cycles) {
        my $tmp_template = $D_TMPELATE;
        $tmp_template =~ s/N/$cycle/;
        $tmp_template =~ s/^T/$time/;

        my ($fh,$fname) = tempfile($tmp_template)
            or croak "can't open tempfile\n";
        push @fileHandles , $fh;
        push @tmpfilenames,$fname;
    }

    @{$r_config->{fileHandles}} = @fileHandles;

    return @tmpfilenames;
}




sub deconv_int {

    my ($deconv,$r_config) = @_;

    my $intensity_file_to_call 
            = $r_config->{'intensity_file_to_call'};

    my $counter=0;
    my %count_to_pos;
    my $number_of_cycles  = $r_config->{'number_of_cycles'};


    if ($intensity_file_to_call =~ m/\.gz$/) {
        $intensity_file_to_call = 
                _convert_gz_to_text_file($intensity_file_to_call);
    }

    open my $FH,'<', $intensity_file_to_call or 
        croak "cant open $intensity_file_to_call: $!\n";

     CLUSTER:
     while (<$FH>) {

        chomp;
        my @list=split /\s+/ , $_;

        my ($lane,$tile,$x,$y)=(shift(@list),
                                shift(@list),
                                shift(@list),
                                shift(@list));
        my $name    = join '_',$lane,$tile,$x,$y;

        $counter++;


        FINDING_Ns:
        foreach my $i (0..scalar @list -1) {
            if ($list[$i] ==0) {
                my $cycle = int ($i / 4) + 1;
                 $count_to_pos{N}{$name}{$cycle} = 1;
            }
        }

        DECONV:
        {
            my $read_matrix = pdl \@list;
            $read_matrix    = 
                        $read_matrix->reshape(4,$number_of_cycles);
            $read_matrix    = 
                        $deconv x $read_matrix;
            $read_matrix    = 
                        $read_matrix->reshape(4*$number_of_cycles,1);
            @list           = 
                        list $read_matrix;
        }

        EXPORT:
        foreach my $cycle (1..$number_of_cycles) {
            my $OUT = ${@{$r_config->{fileHandles}}}[$cycle-1];

            my $intA = shift @list;
            my $intC = shift @list;
            my $intG = shift @list;
            my $intT = shift @list;

            #no intensity detected (Image anal problem:


            my $nuc = '10';
            print {$OUT} 
                    "$nuc"   ,"\t",
                    "1:$intA","$SPACE","2:$intC","$SPACE",
                    "3:$intG","$SPACE","4:$intT","\n"

                or croak "can't print to file\n";
         } #end of export

        REMEMBER:
        {
            $count_to_pos{data}{$counter} = $name;
        }


    }

    if ($r_config->{'verbose'}) {
        print "+Read $counter lines in $intensity_file_to_call\n";
    }

    close $FH;
    return ($counter,\%count_to_pos);
}



sub close_tmp_zone_fh {

    my ($r_config) = @_;
    my @fh_to_close = @{$r_config->{'fileHandles'}};
    foreach my $fh (@fh_to_close) {
        close $fh
            or croak "can't close fileHandles in close_tmp_zone_fh\n";
    }
    delete $r_config->{'fileHandles'};
    return 1;
}


sub arrange_SVM_files {

    my ($r_config) = @_;

    my %SVM_of_cycle;

    my $SVM_package = $r_config->{'svm_files'};

    croak "can't find $SVM_package\n"
        if (not -e $SVM_package);

    my $tar = Archive::Tar->new;
    $tar->read($SVM_package,1);

    my @file_list_in_tar = $tar->list_files;

    if ($r_config->{verbose}) {
        print "+There are ",scalar @file_list_in_tar," files in $SVM_package:\n",
              "@file_list_in_tar\n";
    }

    $tar->extract;

    foreach my $file (@file_list_in_tar) {
        if (not -e $file) {
            croak "Extract failed for $file\n";
        }
    }

    my @model_files = grep {m/model/} @file_list_in_tar;
    my @scale_files = grep {m/range/} @file_list_in_tar;

    foreach my $model_file (@model_files) {
        if ($model_file =~ 
                m/SVM_training_file_for_cycle_(\d{1,3})_/) {
            my $cycle = $1;
            $SVM_of_cycle{'model'}{$cycle} = $model_file;

            if ($r_config->{verbose}) {
                print "+$model_file is model file of cycle $cycle\n";
            } 
        }
        else {
            croak "can't match cycle to: $model_file\n";
        }
    }

    foreach my $scale_file (@scale_files) {
        if ($scale_file =~ 
                m/SVM_training_file_for_cycle_(\d{1,3})_/) {
            my $cycle = $1;
            $SVM_of_cycle{'scale'}{$cycle} = $scale_file;

            if ($r_config->{verbose}) {
                print "+$scale_file is scale file of cycle $cycle\n";
            } 
        }
        else {
            croak "can't match cycle to: $scale_file\n";
        }
    }


    return (\%SVM_of_cycle);
}



sub apply_scaling {

    my ($r_intensity_files_by_cycle,$r_SVM_of_cycle) = @_;
    my @scaled_files;
    my $cycle = 0;

    SCALE:
    foreach my $input_file (@$r_intensity_files_by_cycle) {

        $cycle++;

        my $svm_scale = $r_SVM_of_cycle->{scale}->{$cycle}
            or croak "can't find scale file for cycle $cycle\n";
        my $output_file = $input_file.'.scaled';
        push @scaled_files , $output_file;

        my $command = "$D_SVM_SCALE_COMMAND $D_SVM_SCALE_PARAMS $svm_scale $input_file > $output_file\n";
        print "+",time(),"\t",$command;
        system $command;

    } #end of SCALE

    return (\@scaled_files);
}


sub SVM_them {

    my ($r_scaled_files,$r_SVM_of_cycle,$r_config) = @_;
    my @output_files;
    my $cycle =0;

    my $parameters = $D_SVM_PREDICT_PARAMS;
#     if (defined (my $cross_vald = $r_config->{'cross_vald'})) {
#         $parameters .= $SPACE."-v $cross_vald";
#     }
    if ($r_config->{'prob_estimate'} eq 'Yes') {
        $parameters .= $SPACE."-b 1";
    }

    $parameters .= $SPACE;

    READ:
    foreach my $input_file (@$r_scaled_files) {

        $cycle++;
        my $svm_model = $r_SVM_of_cycle->{model}->{$cycle}
            or croak "can't find model file for cycle $cycle\n";

        my $output = $input_file.'.prediction';
        push @output_files, $output;

        my $command =  "$D_SVM_PREDICT_COMMAND $parameters $input_file $svm_model $output 2>&1";
        print "+",time(), "\tcycle:$cycle<\t",$command,"\n";
        system $command;
    }#end of READ

    return \@output_files;
}


sub convert_to_sequence_and_probability {

    my ($r_output_files, $r_count_to_pos, $r_config) = @_;
    my @seq;
    my @qul;
    my @names;



    my $cycle = 0;
    my $line_count;
    READ_OUTPUT_FILES:
    foreach my $outputfile (@$r_output_files) {
        open my $FH, '<', $outputfile
            or croak "can't open $outputfile: $!\n";

        $cycle++;
        print "$cycle\n";

        $line_count = 0;
        while (<$FH>) {
            chomp;
            next if (not m/^\d/);
            $line_count++;
            my ($nuc, $qA,$qC,$qG,$qT) = split /\s+/ , $_;



            my $name = $r_count_to_pos->{data}->{$line_count};

            $nuc =~ tr/0123/ACGT/;
            if (exists $r_count_to_pos->{N}->{$name}->{$cycle}) {
                $nuc = 'N';
            }

            $name = $r_config->{'flowcell_name'} . $name;


            my @q = ($qA,$qC,$qG,$qT);
            my $qul = maxI (@q);
            $qul -= $EPS;

            my $qul_solexa_style   = -10 * log ((1 - $qul) / $qul) / log (10);
            $qul_solexa_style      = sprintf ("%d",$qul_solexa_style);

            if ($qul_solexa_style > $D_UPPER_LIMIT) {
                $qul_solexa_style = $D_UPPER_LIMIT;
            }

            if (not defined $seq[$line_count-1]) {
                $seq[$line_count-1] = $nuc;
                $qul[$line_count-1] = $qul_solexa_style;
            } else {
                $seq[$line_count-1]    = $seq[$line_count-1] . $nuc;
                $qul[$line_count-1]    = $qul[$line_count-1] . ' ' .$qul_solexa_style;
            }

            $names[$line_count-1]  = $name;
        }
        close $FH;
    } #end of READ_OUTPUT_FILES


    my $sequence_filename  = _generate_filenames($r_config->{intensity_file_to_call});

    open my $OUT, '>', $sequence_filename
        or croak "can't open $sequence_filename for writing\n";

    if ($r_config->{verbose}) {
        print "+Generating sequence file: $sequence_filename\n";
    }

    foreach my $line (0..$line_count-1) {

        print {$OUT} "@",   $names[$line]     ,"\n";
        print {$OUT}        $seq[$line]       ,"\n";
        print {$OUT} "+",   $names[$line]     ,"\n";
        print {$OUT}        "$qul[$line]"     ,"\n";

    }

    if ($r_config->{verbose}) {
        print "+Sequence file has $line_count lines\n";
    }


    close $OUT;
    return $sequence_filename;
}


sub _generate_filenames {

    my ($intensity_filename) = @_;

    my $basename = basename $intensity_filename;
    $basename =~ s/.int.txt//;
    $basename .= $D_SUFFIX_CALLER;
    return $basename
}



 sub send_files_home {
     my ($home_dir, $sequence_filename) = @_;

     if (not -e $sequence_filename) {
        croak "can't find $sequence_filename\n";
     }

    my $fullname_here = abs_path ($sequence_filename);
    my $fullname_there = "$home_dir/$sequence_filename";
    copy ($fullname_here,$fullname_there)
        or croak "could not copy $fullname_here to $fullname_there: $!\n";

    print ">Sequence_at_home:$fullname_there\n";
    unlink $fullname_here;
    return 1;
}




sub _convert_gz_to_text_file {
    my ($gz_file) = @_;
    qx($D_GUNZIP $gz_file);
    $gz_file =~ s/.gz//;
    return $gz_file;
}


sub sumI {
    my @y = @_;
    my $r = 0;
    $r +=$_ foreach @y;
    return $r;

}

sub maxI {
    my @y = @_;
    @y = sort {$a <=> $b} @y;
    my $r = pop @y;
    return $r;
}
