#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)



use strict;
use warnings;
use PDL;
use Readonly;
use Model3;
use List::Util qw (shuffle);
use Storable;
use File::Temp qw (tempfile);
use Carp;
use Getopt::Long;
use Archive::Tar;
use Cwd 'abs_path';
use File::Copy;



Readonly my $SPACE              => q{ };

Readonly my $D_SVM_cost         => '1'; #parameter c
Readonly my $D_SVM_gamma        => undef; #gamma - don't change undef! If you change you shoudl also change to {$gamma = 2 ** $D_SVM_gamma in read_parameters sub}
Readonly my $D_p3               => '0';
Readonly my $D_shrink           => '0';
Readonly my $D_TMPELATE         => 'T_SVM_training_file_for_cycle_N_XXXX';
Readonly my $D_TOKEN_TEMPLATE   => 'token_XXXXX';
Readonly my $D_SVM_SCALE_COMMAND=> 'svm-scale';
Readonly my $D_SVM_SCALE_PARAMS => '-l -1 -u 1';
Readonly my $D_SVM_TRAIN_COMMAND=> 'svm-train ';
Readonly my $D_SVM_TRAIN_PARAMS => '-m 1000';

main();


sub main {

    my $time = time();
    print "+load parameters\n";
    my $r_config = read_parameters();

    print "+reporting parameters\n";
    report_params($r_config);

    print "+load_correction_matrix\n";
    my $deconv = load_correction_matrix ($r_config);

    print "+prepare_tmp_zone_for_training\n";
    my @tmp_training_files = prepare_tmp_zone_for_training($r_config,$time);

    print "+deconvolute intensity file\n";
    my ($counter,$r_pos_to_int) = deconv_int($deconv,$r_config);

         carp "+only $counter lines were collected. Expected:",
         $r_config->{'counter_threshold'},"\n"
            if ($counter < $r_config->{'counter_threshold'});

     print "+close_tmp_zone_fh\n";
     close_tmp_zone_fh($r_config);

     print "+apply_scaling\n";
     my ($r_scaledfiles,$r_range_files) 
                = apply_scaling(\@tmp_training_files);

     print "+SVM_them\n";
     my $r_model_files = SVM_them($r_scaledfiles,$r_config);
    print "finished SVM train\n";

    if (defined $r_config->{bring_files}) {
        print "+Taring files\n";
        send_files_home($r_config->{bring_files},
                        $r_model_files,
                        $r_range_files,
                        $r_config);
    }

}



sub read_parameters {

    #my $self = shift;

    my @random_walk;
    my @svm;
    my $hash_pos_to_seq_name;
    my $unify_int_file;
    my $number_of_cycles;
    my $counter_threshold;
    my $cross_vald;
    my @train_cycles;
    my $prob_estimate;
    my $bring_files;

    my %config;

    GetOptions( "random_walk=f{2}"         => \@random_walk,
                "svm=f{1,2}"               => \@svm,
                "hash_pos_to_seq_file=s"   => \$hash_pos_to_seq_name,
                "number_of_cycles=i"       => \$number_of_cycles,
                "unify_int_file=s"         => \$unify_int_file,
                "counter_threshold=i"      => \$counter_threshold,
                "cross_vald=i"             => \$cross_vald,
                "train_cycles=i{1,}"       => \@train_cycles,
                "prob_estimate"            => \$prob_estimate,
                "bring_files=s"            => \$bring_files,
                );



     my ($p1,$p2) = @random_walk;

     croak "random_walk parameters are not set\n"
         if (not defined $p1 or not defined $p2);

     my ($c,$gamma) = @svm;
     if (defined $c) {$c = 2 ** $c}
     else {$c = 2 ** $D_SVM_cost}

     if (defined $gamma) {$gamma = 2 ** $gamma}
     else {$gamma = $D_SVM_gamma}

     my $r_pos_to_sequence = retrieve ($hash_pos_to_seq_name)
         or croak "can't open hash_pos_to_seq\n";


     croak "number_of_cycles is not set"
         if (not defined $number_of_cycles);

     croak "unify_int_file"
         if (not defined $unify_int_file);

     croak "counter_threshold is not set\n"
         if (not defined $counter_threshold);

     croak "no train_cycles\n"
         if (scalar @train_cycles == 0);


    if ($prob_estimate) {
        $prob_estimate = 'Yes';
    }
    else {
        $prob_estimate = 'No';
    }

    croak "Can't bring files and cross validate in the same time!\n"
    if ((defined $bring_files) and (defined $cross_vald) and ($cross_vald >0));

    croak "Brfing files argument must be absolute path\n"
        if (defined $bring_files and $bring_files=~m/^\./);

    $config{'p1'}                    = $p1;
    $config{'p2'}                    = $p2;
    $config{'c'}                     = $c;
    $config{'gamma'}                 = $gamma;
    $config{'pos_to_seq'}            = $r_pos_to_sequence;
    $config{'number_of_cycles'}      = $number_of_cycles;
    $config{'unify_int_file'}        = $unify_int_file;
    $config{'counter_threshold'}     = $counter_threshold;
    $config{'cross_vald'}            = $cross_vald;
    $config{'train_cycles'}          = \@train_cycles;
    $config{'prob_estimate'}         = $prob_estimate;
    $config{'bring_files'}           = $bring_files;

    return \%config;
}


sub report_params {

    my ($r_config) = @_;
    my @params = qw (p1
                     p2
                     number_of_cycles
                     unify_int_file
                     counter_threshold
                     cross_vald
                     prob_estimate
                     bring_files
                     );

    my $fh = 'STDOUT';
    if (exists $r_config->{token_fh}) {
        #Uglu Code
        $fh = $r_config->{token_fh};
        foreach my $key (@params) {
            my $v = $r_config->{$key} || 'undef';
            print {$fh} ">","$key:","$v","\n";
        }
        my $r_train_cycles = $r_config->{train_cycles};
        print {$fh} ">train_cycles:@$r_train_cycles\n";
        my $c = $r_config->{c};
        my $gamma = $r_config->{gamma};
        $c = log2($c);
        $gamma = log2($gamma);
        print {$fh} ">c:$c\n";
        print {$fh} ">gamma:$gamma\n";
        return 1;
        #End Of Ugly code
    }

    foreach my $key (@params) {

        my $v = $r_config->{$key} || 'undef';
        print ">","$key:","$v","\n";
    }
        my $r_train_cycles = $r_config->{train_cycles};
        print ">train_cycles:@$r_train_cycles\n";
        my $c = $r_config->{c};
        my $gamma = $r_config->{gamma};
        $c = log2($c);
        $gamma = log2($gamma);
        print ">c:$c\n";
        print ">gamma:$gamma\n";
    return 1;
}


sub load_correction_matrix {

    my ($r_config)  =    @_;

    croak "load_correction_matrix error\n"
        if (not defined $r_config);

    my $end_cycle   =    $r_config->{'number_of_cycles'};
    my $p1          =    $r_config->{'p1'};
    my $p2          =    $r_config->{'p2'};

    my $m = Model3->new();
    $m->end_cycle($end_cycle);
    $m->set_parameters($p1,$p2,$D_p3,$D_shrink);
    $m->calculate_matrix();
    my (undef,undef,$deconv_matrix) = $m->inverse_polymerase_matrix();

    return $deconv_matrix;

}

sub prepare_tmp_zone_for_training {

    my ($r_config,$time) = @_;


    my @tmpfilenames,
    my @fileHandles;
    my $number_of_cycles = $r_config->{number_of_cycles};

    foreach my $cycle (1..$number_of_cycles) {
        my $tmp_template = $D_TMPELATE;
        $tmp_template =~ s/N/$cycle/;
        $tmp_template =~ s/^T/$time/;

        my ($fh,$fname) = tempfile($tmp_template)
            or croak "can't open tempfile\n";
        push @fileHandles , $fh;
        push @tmpfilenames,$fname;
    }

    @{$r_config->{fileHandles}} = @fileHandles;

    return @tmpfilenames;
}




sub deconv_int {

    my ($deconv,$r_config) = @_;

    my $unify_int_file = $r_config->{'unify_int_file'};

    my $counter=0;
    my %pos_to_int;
    my $r_pos_to_sequence = $r_config->{'pos_to_seq'};
    my $number_of_cycles  = $r_config->{'number_of_cycles'};
    my $counter_threshold = $r_config->{'counter_threshold'};

    open my $FH,'<', $unify_int_file or 
        croak "cant open $unify_int_file: $!\n";

     CLUSTER:
     while (<$FH>) {

        chomp;
        next if (m/\s0\x2e0\s/);

        my @list=split /\s+/ , $_;

        my ($lane,$tile,$x,$y)=(shift(@list),
                                shift(@list),
                                shift(@list),
                                shift(@list));
        my $name    = join '_',$lane,$tile,$x,$y;
        my $seq     = $r_pos_to_sequence->{$name};

        next if ((not defined $seq)                 or
                 (length $seq < $number_of_cycles)  or
                 ($seq =~ m/N/))                    ;

        $seq       =~ tr/ACGT/0123/;
        $counter++;

        DECONV:
        {
            my $read_matrix = pdl \@list;
            $read_matrix    = 
                        $read_matrix->reshape(4,$number_of_cycles);
            $read_matrix    = 
                        $deconv x $read_matrix;
            $read_matrix    = 
                        $read_matrix->reshape(4*$number_of_cycles,1);
            @list           = 
                        list $read_matrix;
        }

        REMEMBER:
        {
            $pos_to_int{data}{$name}{seq}       = $seq;
            @{ $pos_to_int{data}{$name}{int} }  = @list;
        }

        EXPORT:
        foreach my $cycle (1..$number_of_cycles) {
            my $OUT = ${@{$r_config->{fileHandles}}}[$cycle-1];

            my $intA = shift @list;
            my $intC = shift @list;
            my $intG = shift @list;
            my $intT = shift @list;

            my $nuc = substr $seq, $cycle-1, 1;
            print {$OUT} 
                    "$nuc"   ,"\t",
                    "1:$intA","$SPACE","2:$intC","$SPACE",
                    "3:$intG","$SPACE","4:$intT","\n"

                or croak "can't print to file\n";
         } #end of export

         last if ($counter == $counter_threshold);
    }

    close $FH;
    return ($counter,\%pos_to_int);
}



sub close_tmp_zone_fh {

    my ($r_config) = @_;
    my @fh_to_close = @{$r_config->{'fileHandles'}};
    foreach my $fh (@fh_to_close) {
        close $fh
            or croak "can't close fileHandles in close_tmp_zone_fh\n";
    }
    return 1;
}


sub apply_scaling {

    my ($r_training_cycles_files) = @_;
    my @scaled_files;
    my @range_files;

    SCALE:
    foreach my $input_file (@$r_training_cycles_files) {

        my $output_file = $input_file.'.scaled';
        my $range_file  = $input_file.'.range';

        push @scaled_files , $output_file;
        push @range_files  , $range_file;

        print "+",time(),
              "\t","$D_SVM_SCALE_COMMAND $D_SVM_SCALE_PARAMS -s $range_file $input_file > $output_file\n";
        system ("$D_SVM_SCALE_COMMAND $D_SVM_SCALE_PARAMS -s $range_file $input_file > $output_file");

    } #end of SCALE

    return (\@scaled_files,\@range_files);
}



sub SVM_them {

    my ($r_train_files,$r_config) = @_;

    my @model_files;

    my $number_of_cycles  =
            $r_config->{'number_of_cycles'};
    my $r_train_cycles    =
            $r_config->{'train_cycles'};

    my $parameters = $D_SVM_TRAIN_PARAMS;


    if (defined (my $cross_vald = $r_config->{'cross_vald'})) {
        $parameters .= $SPACE."-v $cross_vald";
    }

    if (defined (my $c = $r_config->{'c'})) {
        $parameters .= $SPACE."-c $c";
    }

    if (defined (my $gamma = $r_config->{'gamma'})) {
        $parameters .= $SPACE."-g $gamma";
    }

    if ($r_config->{'prob_estimate'} eq 'Yes') {
        $parameters .= $SPACE."-b 1";
    }

    $parameters .= $SPACE;

    foreach my $train_cycle (@$r_train_cycles) {
        my $train_file = $r_train_files->[$train_cycle-1];
        my $output     = $train_file . '.model';

        push @model_files, $output;
        print '>',time(),
              "\t",">cycle:$train_cycle<\t",
              "$D_SVM_TRAIN_COMMAND $parameters $train_file $output 2>&1\n";

        system "$D_SVM_TRAIN_COMMAND $parameters $train_file $output 2>&1";
    }

    return \@model_files;

}

sub send_files_home {
    my ($home_dir, $r_model_files, $r_scale_files,$r_config) = @_;

    my $token_name = _put_token($r_config);
     my @allFiles = (@$r_model_files,@$r_scale_files,$token_name);
     foreach my $file (@allFiles) {
         if (not -e $file) {
             croak "can't find $file\n";
         }
     }

    my $tar = Archive::Tar->new;
    $tar->add_files(@allFiles);

    my $filename = $ENV{'JOB_ID'} .       '_' .
                   $ENV{'SGE_TASK_ID'}.   '_' .
                   'returnedFiles.tar.gz';

    $tar->write($filename,1);
    my $fullname_here = abs_path ($filename);
    my $fullname_there = "$home_dir/$filename";
    copy ($fullname_here,$fullname_there)
        or croak "could not copy $fullname_here to $fullname_there: $!\n";

    print ">Tar_at_home:$fullname_there\n";
    unlink $token_name ||
        croak "can't clean token\n";
    return 1;
}




sub _put_token {
    my ($r_config) = @_;
    my ($fh,$fname) = tempfile($D_TOKEN_TEMPLATE);
    $r_config->{token_fh} = $fh;
    report_params($r_config);
    close $fh;
    return $fname;
}

sub log2 {
    my ($number) = @_;
    my $r = (log ($number) / log(2));
    return $r;
}

