#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

# This package submits jobs to the cluster and manages them.
# It is heavily relied on the textual format of the status command.
# The main method is automatic_pilot that manages everything.


package Submit_tools;

use strict;
use warnings;
use Readonly;
use Carp;
use Cwd;
use File::Temp qw(tempfile);
use Time::Local;
use ExtraUtils qw(timestamp_file);
use List::MoreUtils qw (any);
use POSIX qw(ceil);
use File::Basename;
use List::Util qw (max);
use Term::ReadKey;
use File::Glob ':glob';
use Term::ANSIColor qw(:constants);

Readonly my $D_SUBMIT_COMMAND   => 'qsub'                               ;
Readonly my $D_DEL_COMMAND      => 'qdel'                               ;
Readonly my $D_STATUS_COMMAND   => 'qstat'                              ;
Readonly my $D_SH_TEMPLATE      => 'tmp_XXXX'                           ;
Readonly my $D_MEMORY_REQUEST   => '-l virtual_free'                    ;
Readonly my $D_LIFETIME         =>  '1h'                                ;
Readonly my $D_MAX_SIZE_JOBS    =>  '1000000'                           ;
Readonly my $D_OUTDIR           =>  'Cluster'                           ;
Readonly my $D_STDOUT_DIR         =>  'Messages'                        ;
Readonly my $D_STDERR_DIR         =>  'Errors'                          ;

$Term::ANSIColor::AUTORESET = 1;




sub test{
    use File::Copy;
    my $file = shift || die "no file in test\n";
    my $filebase = basename($file) . '.1';
    copy ($file, $filebase) 
        or die $!;
    chmod oct '0777', './sge_script_2008_04_15_20_39_52.pl' 
        or die $!;
     my $sge = Submit_tools->new();
     $sge->range(1,1000,1);
     $sge->submit('/home2/hannon/erlich/SGE-API/step0/Cluster/sge_script_2008_04_15_20_39_52.pl');
     $sge->automatic_pilot('0.1m');
}


sub new {
        #Create a new object
	my $classname = shift;
	my $self = {};
	bless $self,$classname;

        $self->allocate_filesystem_space;
        $self->life_time($D_LIFETIME);
	return $self;
}


sub allocate_filesystem_space {

    my $self = shift;
    if (not -d $D_OUTDIR) {
        mkdir "$D_OUTDIR" or confess "Can't mkdir $D_OUTDIR: $!";
    }
    if (not -d "$D_OUTDIR/$D_STDOUT_DIR") {
        mkdir "$D_OUTDIR/$D_STDOUT_DIR" or confess "Can't mkdir $D_STDOUT_DIR: $!";
    }
    if (not -d "$D_OUTDIR/$D_STDERR_DIR") {
        mkdir "$D_OUTDIR/$D_STDERR_DIR" or confess "Can't mkdir $D_STDERR_DIR: $!";
    }
    $self->{meta}->{error_path}   = "$D_OUTDIR/$D_STDERR_DIR";
    $self->{meta}->{message_path} = "$D_OUTDIR/$D_STDOUT_DIR";

}



################################# Pilot ############################

sub automatic_pilot {



    my ($self,$pilot_refresh_time_m) = @_;
    $self->set_pilot_refresh_time($pilot_refresh_time_m);

    my $counter = $self->number_of_tasks;
    my $alive_tasks = 0;
    my $pilot_refresh_time_s = $self->get_pilot_refresh_time; #seconds!
    my $last_time = time();
    my $job_id = $self->get_job_id;

    FLY:do {
        $alive_tasks = $self->clean;
        my $time = _display_time();
        my $finished = ceil ((1-($alive_tasks/$counter))*100) ;
        my $max_time_to_end = $self->max_time_span;

        print "$time\t",
              "Tasks alive: $alive_tasks out of $counter. ",
              "Finished: $finished%\t",
              "Max lifetime: $max_time_to_end","s\n";
         if ($alive_tasks) {
            while (time()-$last_time < $pilot_refresh_time_s) {
            ReadMode ('cbreak');
                my $chr;
                if (defined ($chr = ReadKey(-1)) and ($chr eq 'q')) {
                    $self->set_job_status($job_id,'finished');
                    $self->delete_job($job_id);
                    ReadMode ('normal');
                    exit(0);
                }
                ReadMode ('normal');
            }
        }
        $last_time = time();
    } while ($alive_tasks);
    return 1;
}


sub _display_time {
    my ($sec,$min,$hour,$day,$mon,$year) = localtime(time);
    $year  += 1900;
    $mon += 1;
    return sprintf("%04d/%02d/%02d at %02d:%02d:%02d", $year,$mon,$day,$hour,$min,$sec);
}


sub get_pilot_refresh_time {
    my $self = shift;
    return $self->{meta}->{pilot_refresh};
}

sub set_pilot_refresh_time {

    my $self = shift;
    if (@_) {
        my $update_time = shift;
        $update_time =~ s/m//i;
        $self->{meta}->{pilot_refresh} = $update_time * 60;
        return $self->{meta}->{pilot_refresh};
    }
    return;
}


################################# Submit ############################

sub submit {


    my $self = shift;
    my $filename = shift;


    if (not -e $filename) {
        confess "can't find filename\n";
    }

    my ($fh,$sh_file) = tempfile($D_SH_TEMPLATE, SUFFIX => '.sh');

    my ($start,$stop,$gap) = $self->range;
    confess "can't submit before setting range\n" 
        if any {not defined $_} $start,$stop,$gap;

    my $perl_file = $filename;
    my $perl_file_no_path = basename $perl_file;

    chmod oct '0777',$sh_file;
    print {$fh} "#\$ -S /bin/bash                    \n",
                "#\$ -cwd                            \n",
                "#\$ -v PATH                         \n",
                                                    "\n",
                "#\$ -t $start-$stop:$gap            \n",
                                                    "\n",
                "SDIR=\$PWD                          \n",
                "cd \$TMPDIR                         \n",
                "cp $perl_file .                     \n",
                "./$perl_file_no_path \$SGE_TASK_ID  \n",

    ;



    my $memory_request = " ";
    if (exists $self->{meta}->{memory}) {
        my $memory = $self->{meta}->{memory};
        $memory_request = "$D_MEMORY_REQUEST=$memory";
    }
    my $error_path   = $self->error_path;
    my $message_path =  $self->message_path;


    my $submit_response = `$D_SUBMIT_COMMAND $memory_request -o $message_path -e $error_path $sh_file 2>&1`;
    $submit_response =~ m/Your job-array (\d+)\..*has been submitted/;
    confess "submit error: $submit_response\n"
        if (not defined $1);
    my $job_id = $1;
    print _display_time,"\t","Starting $job_id\n";
    close $fh;


    $self->set_job       ($job_id);
    $self->set_job_status($job_id,'requested');
    $self->perl_file     ($job_id,$perl_file);
    $self->sh_file       ($job_id,$sh_file);
    $self->set_message_filenames;
    $self->set_error_filenames;



    foreach my $task (@{$self->get_task_array}) {
        $self->set_task_status($job_id,$task,'requested');
    }

    return $job_id;
}


################################# Update times ############################
sub set_task_update_time {
    my $self = shift;
    my ($job_id,$r_task) = @_;
    confess "set_task_time error\n"
        if any {not defined $_} $job_id, $r_task;

    if (ref $r_task eq 'ARRAY') {foreach (@$r_task) {$self->{data}->{$job_id}->{$_}->{time}->{updated} = time()}}
    else {$self->{data}->{$job_id}->{$r_task}->{time}->{updated} = time()}

    return 1;
}
sub get_task_update_time {
    my $self = shift;
    my ($job_id,$task) = @_;
    confess "get_task_time error\n"
        if any {not defined $_} $job_id, $task;
    return $self->{data}->{$job_id}->{$task}->{time}->{updated};
}
sub set_job_update_time {
    my $self = shift;
    my ($job_id) = @_;
    confess "set_job_update_time error\n"
        if any {not defined $_} $job_id;
    $self->{data}->{$job_id}->{time}->{updated};
}
sub get_job_update_time {
    my $self = shift;
    my ($job_id,$task) = @_;
    confess "get_job_update_time error\n"
        if any {not defined $_} $job_id;
    return $self->{data}->{$job_id}->{time}->{updated}
}

################################# Status ############################
sub set_task_status {
    my ($self,$job_id,$r_task,$status) = @_;
    confess "set_task_status error\n"
        if any {not defined $_} $job_id, $r_task, $status;

    if (ref $r_task eq 'ARRAY') {foreach (@$r_task) {$self->{data}->{$job_id}->{$_}->{status} = $status}}
    else {$self->{data}->{$job_id}->{$r_task}->{status} = $status}

    $self->set_task_update_time($job_id,$r_task);

    return 1;
}
sub get_task_status {
    my $self = shift;
    my ($job_id,$task) = @_;
    confess "get_task_status error\n"
        if any {not defined $_} $job_id, $task;

    return ($self->{data}->{$job_id}->{$task}->{status},
            $self->get_task_update_time($job_id,$task));
}

sub set_job_status {
    my $self = shift;
    my ($job_id,$status) = @_;
    confess "set_job_status error\n"
        if any {not defined $_} $job_id, $status;
    $self->{data}->{$job_id}->{status} = $status;

    $self->set_job_update_time($job_id);

    return $self->{data}->{$job_id}->{status};
}
sub get_job_status {
   my $self = shift;
   my ($job_id) = @_;
   confess "get_job_status error\n"
       if any {not defined $_} $job_id;

   return ($self->{data}->{$job_id}->{status},
          $self->get_job_update_time($job_id));
}

################################# Files ############################
sub perl_file {
    my $self = shift;
    my $job_id = shift || confess "perl_file error\n";
    if (@_) {
        my $file = shift;
        $self->{data}->{$job_id}->{perl_file} = $file;
   }
    return $self->{data}->{$job_id}->{perl_file};
}
sub sh_file {
    my $self = shift;
    my $job_id = shift || confess "sh_file error\n";
    if (@_) {
        my $file = shift;
        $self->{data}->{$job_id}->{sh_file} = $file;
   }
    return $self->{data}->{$job_id}->{sh_file};
}

################################# Meta ############################
sub set_job {
    my $self = shift;
    my ($job_id) = @_;
    confess "set_job error\n"
        if any {not defined $_} $job_id;

    $self->{meta}->{job} = $job_id;
}

sub get_job_id {
    my $self = shift;
    return $self->{meta}->{job}
}
################################# Range ############################
sub range {

    my $self = shift;
    if (@_) {
        my ($first,$last,$gap) = @_;
        confess "range error" 
            if any {not defined $_} $first,$last;
        $gap = 1 if (not defined $gap);
        my @array;

        $self->{meta}->{range}->{start} = $first;
        $self->{meta}->{range}->{stop}  = $last;
        $self->{meta}->{range}->{gap}  =  $gap;

        for (my $i = $first; $i<= $last; $i+= $gap) {
            push @array, $i;
        }

        @{$self->{meta}->{range}->{array}}  =  @array;

        carp "range is very big!\n" 
           if (scalar @array > $D_MAX_SIZE_JOBS);
    }

    return ($self->{meta}->{range}->{start},
            $self->{meta}->{range}->{stop} ,
            $self->{meta}->{range}->{gap});

}

sub get_task_array {
    my $self = shift;
    return \@{$self->{meta}->{range}->{array}};
}

sub set_task_array {
    my $self = shift;
    if (@_) {
        my $r_a = shift;
        @{ $self->{meta}->{range}->{array} } = @{$r_a};
        return \@{ $self->{meta}->{range}->{array} };
    }
    return;
}

################################# Update_Status ############################

sub update_status {

    my $self = shift;
    my $job_id = $self->get_job_id;
    my $r_all_tasks = $self->get_task_array;

    my %h_all_tasks; 
    $h_all_tasks{$_}++ foreach (@$r_all_tasks);

    my $total_tasks = $self->number_of_tasks;
    my $time_now   = time();
    my $flag_error = 0;



    my $status_response = `$D_STATUS_COMMAND`;
    my %job_id_status = %{_parse_update($status_response)};


    if (exists $job_id_status{$job_id}) {
        UPDATE_CYCLE: foreach my $task_reported (keys %{$job_id_status{$job_id}}) {

            my $r_task_array = _extract_array($task_reported);
            delete $h_all_tasks{$_} foreach (@{$r_task_array});

            $self->set_task_status($job_id,
                                   $r_task_array,
                                   $job_id_status{$job_id}{$task_reported}{'status'}
                                   );

            $self->set_born($job_id,
                              $r_task_array,
                              $job_id_status{$job_id}{$task_reported}{'start_time'}
                             );

            ERORR: if ($job_id_status{$job_id}{$task_reported}{'status'} =~ m/E/) {
                carp "problem in $job_id $task_reported\n";
                $flag_error = 1;
            } #end of error
        } #end of UPDATE_CYCLE
    }


    my @keys = keys %h_all_tasks;
    FINISHED_JOBS: if (scalar @keys) {
        $self->set_task_status($job_id,\@keys,'finished');
    }

    my $n_active_tasks = $total_tasks - scalar keys %h_all_tasks;
    #     print "active: $n_active_tasks\n";
    #     print "total: $total_tasks\n";
    #     print "finished:", scalar @keys,"\n";

    if ($flag_error == 0) {
        return ($n_active_tasks,$total_tasks);
    } else {
       return;
    }
}

sub _parse_update {

    my ($status_response) = @_;
    my %job_id_status;

    my @status_response_lines = split /\n/,$status_response;
    @status_response_lines = grep {/^\d/} @status_response_lines;

    foreach my $line (@status_response_lines) {
        my @fields = split /\s+/, $line;
        my ($job_id,undef,undef,undef,$status,$date,$time) = @fields;
        my $task_id = pop @fields; #the number of fields 
                                  #is changing sometimes when it goes fro qw to r 
                                  # but task is always the last one
        my ($month, $day, $year) = split /\//, $date;
        my ($hour,$min,$sec)     = split /:/ , $time;


        $month -=1; #adapt it to 0..11 format
        $year -= 1900;#adapt it to year format


        my $epoch_start = timelocal($sec,$min,$hour,$day,$month,$year);
        $job_id_status{$job_id}{$task_id}{'status'} = $status;
        $job_id_status{$job_id}{$task_id}{'start_time'} = $epoch_start;



        #print "$line\n";
        #print "$job_id,undef,undef,undef,$status,$date,time:$time,q:$queue,t:$task_id\n";

    }


    return \%job_id_status;
}


sub _extract_array {
    my ($task_reported) = @_;
    my @a;
    if ($task_reported =~ m/(\d*)-(\d*):(\d*)/) {
        my ($start,$stop,$gap) = ($1,$2,$3);
        for (my $i = $start;$i <= $stop; $i+=$gap) {push @a,$i};
    }
    elsif ($task_reported =~ m/\d+,\d+/) {
        @a = split /,/ , $task_reported;
    }

    else {
        push @a, $task_reported;
    }
    return \@a;
}

################################# Clean ############################
sub clean {

    my $self = shift;
    my $job_id = $self->get_job_id;
    my $r_tasks = $self->get_task_array;

    my ($active_tasks,$total_tasks) = $self->update_status;
    if (not defined $active_tasks) {carp "error in $job_id\n"};

    foreach my $task (@$r_tasks) {

        if ($self->is_finished($job_id,$task)) {
            $self->delete_task($job_id,$task)}

        else {
              if ($self->task_is_dead($job_id,$task)){
                $self->delete_task($job_id,$task) 
              }
       }

    }

    if ($self->number_of_tasks == 0) {
        $self->set_job_status($job_id,'finished');
        $self->delete_job($job_id);
    }
    return $self->number_of_tasks;
}

################################# Delete ############################

sub delete_task {
    my ($self,$job_id,$task) = @_;
    confess "delete_task error\n"
        if any {not defined $_} $job_id,$task;

    my %h;
    $self->clean_error_file($task);
    $h{$_}++ foreach (@{$self->get_task_array});
    delete $h{$task};
    delete $self->{data}->{$job_id}->{$task};
    my @keys = keys %h;
    $self->set_task_array(\@keys);
}


sub delete_job {
    my ($self,$job_id) = @_;
    confess "delete_job error\n"
        if (not defined $job_id);
    my $del_response = `$D_DEL_COMMAND $job_id 2>&1`;
    unlink $self->perl_file($job_id)
        or carp "problem clean perl file of $job_id\n";
    unlink $self->sh_file($job_id)
        or carp "problem clean sh file of $job_id\n";

   if (     (not $del_response =~ m/has deleted job $job_id/)
        and (not $self->job_is_finished($job_id)) ) {

        carp "can't delete job $job_id: $del_response\n";
        return 
    }

    delete $self->{data}->{$job_id};
    return 1;
}

################################# Task Activity ############################
sub is_finished {

    my ($self,$job_id,$task) = @_;
    confess "is_finished function error\n"
        if any {not defined $_} $job_id, $task;

    my ($status,undef) = $self->get_task_status($job_id,$task);

    if ($status eq 'finished') {
        return 1;
    }
    return ;
}

sub job_is_finished {
    my ($self,$job_id) = @_;
    confess "job_is_finished error\n"
        if any {not defined $_} $job_id;
    my ($status,undef) = $self->get_job_status($job_id);
    if ($status eq 'finished') {
        return 1;
    }
    return;
}


sub number_of_tasks {
    my $self = shift;
    return scalar @{($self->get_task_array)};
}

################################# Life Time ############################

sub get_born {
    my $self = shift;
    my ($job_id,$task) = @_;
    confess "get_born error\n"
        if any {not defined $_} $job_id,$task;
        return $self->{data}->{$job_id}->{$task}->{time}->{born};
}

sub set_born {
   my $self = shift;
   my ($job_id, $r_task,$time) = @_;
   confess "set_born error\n"
        if any {not defined $_} $job_id, $r_task, $time;


   if (ref $r_task eq 'ARRAY') {foreach (@$r_task){$self->{data}->{$job_id}->{$_}->{time}->{born} = $time}}
   else {$self->{data}->{$job_id}->{$r_task}->{time}->{born} = $time}


    return 1;
}

sub life_time {
    my ($self,$lifetime_hour) = @_;
    if (defined $lifetime_hour) {
        $lifetime_hour =~ s/h//i;
        my $lifetime_sec = $lifetime_hour * 3600;
        $self->{meta}->{life_time} = $lifetime_sec;
    }
    return $self->{meta}->{life_time};
}

sub task_is_dead {

    my ($self,$job_id,$task) = @_;
    confess "task_is_dead error\n"
        if any {not defined $_}  $job_id,$task;

    my $time_span = time() - $self->get_born($job_id,$task);
    my $life      = $self->life_time;

#     print "time:",time(), "\t",
#           "xy:"  ,$self->get_born($job_id,$task),"\t",
#           "Life:",$self->life_time,"\t",
#           "timespan:",$time_span;
#           "Tnai:\n";
# 
    if ($time_span >= $life) {
        return 1
    } 
    return;
}

sub max_time_span {
    my $self = shift;
    my @timespan;
    my $job_id = $self->get_job_id;
    foreach my $task (@{$self->get_task_array}) {
        my $d = 0;
        if (defined( $self->life_time) and defined($self->get_born($job_id,$task))) {
            $d = $self->life_time - (time() - $self->get_born($job_id,$task));
        }
        push @timespan,$d;
    }
    my $max_timesapen = max @timespan;
    if ((not defined $max_timesapen) or ($max_timesapen < 0)) {$max_timesapen = 0};
    return $max_timesapen; 
}


################################# Parameters ############################

sub message_path {
   my $self = shift;
    if (@_) {
        $self->{meta}->{message_path} = shift;
    }
    return $self->{meta}->{message_path}; 
}
sub set_message_filenames {
    my $self = shift;
    my $path_to_look = $self->message_path;
    confess "no path to look\n"
        if (not $path_to_look);

    my $job_id = $self->get_job_id;
    my $sh_file = $self->sh_file($job_id);
    my $r_task_array = $self->get_task_array;
    my @filenames;
    foreach my $task (@$r_task_array) {
        my $filename = "$path_to_look/$sh_file.o$job_id.$task";
        push @filenames, $filename;
    }
    @{$self->{data}->{message_filenames}} = @filenames;
    return @{$self->{data}->{message_filenames}};
}
sub get_message_filenames {
    my $self = shift;
    return @{$self->{data}->{message_filenames}};
}

sub set_error_filenames {
    my $self = shift;
    my $path_to_look = $self->error_path;
    confess "no path to look\n"
        if (not $path_to_look);

    my $job_id = $self->get_job_id;
    my $sh_file = $self->sh_file($job_id);
    my $r_task_array = $self->get_task_array;
    my @filenames;
    foreach my $task (@$r_task_array) {
        my $filename = "$path_to_look/"."$sh_file.e$job_id.$task";
        push @filenames, $filename;
    }
    @{$self->{data}->{error_filenames}} = @filenames;
    return @{$self->{data}->{error_filenames}};
}

sub get_error_filenames {
    my $self = shift;
    return @{$self->{data}->{error_filenames}};
}

sub get_error_filename_by_task {
    my $self = shift;
    my $task = shift || confess "error\n";
    my $job_id = $self->get_job_id;
    my $sh_file = $self->sh_file($job_id);
    my $filename = "$sh_file.e$job_id.$task";

    $filename = $self->error_path . '/'. $filename;

    return $filename;
}

sub clean_error_file {
    my $self = shift;
    my $task = shift || confess "error\n";

    my $file_to_clean = $self->get_error_filename_by_task($task);
    open my $ER,'<',$file_to_clean
        or carp "can't open error file: $file_to_clean of task $task: $!\n";

    print "\nContent of $file_to_clean:\n";
    while (<$ER>) {print BOLD, RED "$_";}
    print "$. messages\n";
    print "Cleaining file\n";
    print "\n";
    unlink $file_to_clean or confess "can't clean: $file_to_clean of task $task: $!\n";
    close $ER;

    return 1;
}


sub error_path {
   my $self = shift;
    if (@_) {
        $self->{meta}->{error_path} = shift;
    }
    return $self->{meta}->{error_path}; 
}



sub memory {

    my $self = shift;
    if (@_) {
        $self->{meta}->{memory} = shift;
    }
    return $self->{meta}->{memory};
}
1;
